
#include "SPDMean.h"

/*Define the namespace*/
namespace ROPTLIB{

	SPDMean::SPDMean(double *inLs, double *inXtrue, integer inn, integer innum)
	{
		Ls = inLs;
        Xtrue = inXtrue;
		n = inn;
		num = innum;
	};

	SPDMean::~SPDMean(void)
	{
	};

	double SPDMean::f(Variable *x) const
	{
		SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
		if (!x->TempDataExist("L"))
		{
			Mani->CholeskyRepresentation(x);
		}
		const SharedSpace *SharedL = x->ObtainReadTempData("L");
		Variable *LElement = SharedL->GetSharedElement();
		const double *L = LElement->ObtainReadData();

		SharedSpace *SharedlogLXL = new SharedSpace(3, n, n, num);
		double *logLXL = SharedlogLXL->ObtainWriteEntireData();
		double *Ltmp = new double[n * n];
		integer length = n * n, N = n, info;
		for (integer i = 0; i < num; i++)
		{
			dcopy_(&length, const_cast<double *> (L), &GLOBAL::IONE, Ltmp, &GLOBAL::IONE);
			/*Solve the linear system Li X = Lx, i.e., X = Li^{-1} Lx. The solution X is stored in LiiLx.
			Note that Li is a lower triangular matrix.
			Details: http://www.netlib.org/lapack/explore-html/d6/d6f/dtrtrs_8f.html */
			dtrtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, Ls + n * n * i, &N, Ltmp, &N, &info);
			if (info != 0)
			{
				std::cout << "The cholesky decompsotion in SPDMean::f failed with info:" << info << "!" << std::endl;
			}
			dgemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, Ltmp, &N, Ltmp, &N, &GLOBAL::DZERO, logLXL + n * n * i, &N);
			Matrix MMt(logLXL + n * n * i, n, n);
			Matrix::LogSymmetricM(GLOBAL::L, MMt, MMt);
		}
		length = n * n * num;

		delete[] Ltmp;
		x->AddToTempData("logLXL", SharedlogLXL);

		double result = dnrm2_(&length, logLXL, &GLOBAL::IONE);
		result *= result;
		result /= 2.0 * num;
		return result;
	};
    
    

	void SPDMean::RieGrad(Variable *x, Vector *gf) const
	{
		const SharedSpace *SharedlogLXL = x->ObtainReadTempData("logLXL");
		const double *logLXL = SharedlogLXL->ObtainReadData();
		double *gfVT = gf->ObtainWriteEntireData();
		for (integer i = 0; i < n * n; i++)
			gfVT[i] = 0;
		const double *xM = x->ObtainReadData();

		double *tmp = new double[n * n];
		integer N = n, info;
		for (integer i = 0; i < num; i++)
		{
			/*tmp <-- log(Li^{-1} X Li^{-T}) Li^T */
			dgemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, const_cast<double *> (logLXL + n * n * i), &N,
				Ls + n * n * i, &N, &GLOBAL::DZERO, tmp, &N);

			/*Solve the linear system Li^T X = tmp, i.e., X = Li^{-T} log(Li^{-1} X Li^{-T}) Li^T. The solution X is stored in tmp.
			Note that Li is a lower triangular matrix.
			Details: http://www.netlib.org/lapack/explore-html/d6/d6f/dtrtrs_8f.html */
			dtrtrs_(GLOBAL::L, GLOBAL::T, GLOBAL::N, &N, &N, Ls + n * n * i, &N, tmp, &N, &info);
			if (info != 0)
			{
				std::cout << "The cholesky decompsotion in SPDMean::RieGrad failed with info:" << info << "!" << std::endl;
			}
			dgemm_(GLOBAL::N, GLOBAL::N, &N, &N, &N, &GLOBAL::DONE, const_cast<double *> (xM), &N, tmp, &N, &GLOBAL::DONE, gfVT, &N);
		}
		double scalar = 1.0 / num;
		integer length = n * n;
		dscal_(&length, &scalar, gfVT, &GLOBAL::IONE);
		delete[] tmp;
	};
    

    
    // Compute the Riemannian distance between x and Xtrue(i.e., the provided true solution)
    double SPDMean::dis(Variable *x) const
    {
         SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
        // dis(x, y) = |log(Lx^{-1} y Lx^{-T})|_F
        if (!x->TempDataExist("L"))
		{
			Mani->CholeskyRepresentation(x);
		}
		const SharedSpace *SharedLx = x->ObtainReadTempData("L");
		Variable *LElementx = SharedLx->GetSharedElement();
		const double *Lx = LElementx->ObtainReadData();
         
        integer info, N = n, length = n * n;
        double *y = new double[n * n];
        dcopy_(&length, Xtrue, &GLOBAL::IONE, y, &GLOBAL::IONE);
         
        // Cholesky factorization
        dpotrf_(GLOBAL::L, &N, y, &N, &info);
        for (integer j = 0; j < n; j++)
        {
            for (integer k = j + 1; k < n; k++)
            {
                y[j + k * n] = 0; 
            }
        }

		double *LxLy = new double[n * n];
         
		dcopy_(&length, y, &GLOBAL::IONE, LxLy, &GLOBAL::IONE);
 		/*Solve the linear system Lx X = Ly, i.e., X = Lx^{-1}Ly. The solution X is stored in LxLy*/
        dtrtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<double *> (Lx), &N, LxLy, &N, &info);
         
         /* compute temp = LxLy*(LxLy)^T  */
        double *temp = new double[n * n];
        dgemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, LxLy, &N, LxLy, &N, &GLOBAL::DZERO, temp, &N);
         
        double *eigenvalues = new double[n + n * n];
	    double *eigenvectors = eigenvalues + n;
        Matrix E(eigenvalues, n, 1), V(eigenvectors, n, n);
        Matrix MMt(temp, n, n);
	    Matrix::EigenSymmetricM(GLOBAL::L, MMt, E, V);
        
        double result = 0.0;
        for (integer i = 0; i < N; i++) 
        {
            result = result + log(eigenvalues[i])*log(eigenvalues[i]);
        }

        result = std::sqrt(result);        
        delete[] eigenvalues;   
        delete[] temp;
        delete[] y;
        delete[] LxLy;
        return result;
       
    };
   
    
    
    
    double SPDMean::distwo(Variable *x, Variable *y) const
    {
        // Compute the affine invariant distance between x and y
        // dis(x, y) = |log(Lx^{-1} y Lx^{-T})|_F
        
        SPDManifold *Mani = dynamic_cast<SPDManifold *> (Domain);
        if (!x->TempDataExist("L"))
		{
			Mani->CholeskyRepresentation(x);
		}
		const SharedSpace *SharedLx = x->ObtainReadTempData("L");
		Variable *LElementx = SharedLx->GetSharedElement();
		const double *Lx = LElementx->ObtainReadData();
        
        if (!y->TempDataExist("L"))
		{
			Mani->CholeskyRepresentation(y);
		}
		const SharedSpace *SharedLy = y->ObtainReadTempData("L");
		Variable *LElementy = SharedLy->GetSharedElement();
		const double *Ly = LElementy->ObtainReadData();
        
        integer N = n, info, length = n * n;
		double *LxLy = new double[N * N];
        
		dcopy_(&length, const_cast<double *> (Ly), &GLOBAL::IONE, LxLy, &GLOBAL::IONE);
		/*Solve the linear system Lx X = Ly, i.e., X = Lx^{-1}Ly. The solution X is stored in LxLy*/
        dtrtrs_(GLOBAL::L, GLOBAL::N, GLOBAL::N, &N, &N, const_cast<double *> (Lx), &N, LxLy, &N, &info);
        
        /* compute temp = LxLy*(LxLy)^T  */
        double *temp = new double[N * N];
        dgemm_(GLOBAL::N, GLOBAL::T, &N, &N, &N, &GLOBAL::DONE, LxLy, &N, LxLy, &N, &GLOBAL::DZERO, temp, &N);
        
        double *eigenvalues = new double[n + n * n];
	    double *eigenvectors = eigenvalues + n;
        Matrix E(eigenvalues, n, 1), V(eigenvectors, n, n);
        Matrix MMt(temp, n, n);
	    Matrix::EigenSymmetricM(GLOBAL::L, MMt, E, V);
        
        double result = 0.0;
        for (integer i = 0; i < N; i++) 
        {
            result = result + log(eigenvalues[i])*log(eigenvalues[i]);
        }
         
        result = sqrt(result);
        
        delete[] eigenvalues;   
        delete[] temp;
        delete[] LxLy;
        return result;
        
    }
    
    

	void SPDMean::RieHessianEta(Variable *x, Vector *etax, Vector *xix) const
	{
		std::cout << "warning: SPDMean::RieHessianEta has not been implemented!" << std::endl;
		etax->CopyTo(xix);
	};
}; /*end of ROPTLIB namespace*/
