function [SPDMean, f, gf, gfgf0, iter, nf, ng, nR, nV, nVp, nH, ComTime, funs, grads, times, diss, initstepsizes, stepsizes, truestepsizes, relfs] = ComputeSPDKarcherMean(As, Xinitial, Xtrue, SolverParams)
% This function is used to compute the Karcher mean 

% input: 
    % As: data points
    % Xinitial: initial iterate
    % Xtrue: true solution (Xtrue is needed if one wants to compute the
    %        true error--the distance between current iterate and the exact
    %        solution--at each iteration. Otherwise make it an identity
    %        matrix or any other SPD matrix)
    % SolverParams: parameters

% output:
    % SPDMean: unique minimizer
    % f: function value
    % gf: norm of gradient
    % gfgf0: relative norm of gradient gf/gf0
    % iter: number of iterations
    % nf: number of function evaluations
    % ng: number of gradient evaluations
    % nR: number of retraction evaluations
    % nV/nVp: number of actions of vector transport
    % nH: number of evaluations of the action of Hessian
    % ComTime: total computational time (seconds)
    % funs: function values at each iterate
    % grads: norm of gradient at each iterate
    % times: computational time for each iterate
    % diss: Riemannian distance between current iterate and the true
    %       solution if the true solution is given
    % initstepsizes: initial stepsize at each iterate
    % stepsizes: final stepsize at each iterate
    % truestepsizes: Riemannian distance between current iterate and
    %                previous iterate
    % relfs: df/f

    n = size(As{1}, 1);
    k = length(As);
    Ls = zeros(n, n, k);
    
    % compute the Cholesky factorization of Ai's
    for i = 1 : k
        Ls(:, :, i) = chol(As{i}, 'lower');
    end
    
    % indicate if the locking condition is satisfied
    if(~isfield(SolverParams, 'HasHHR'))
        HasHHR = 1;
    else
        HasHHR = SolverParams.HasHHR;
    end
    
    % if there is no input method, take the LRBFGS as default
    if(~isfield(SolverParams, 'method'))
        SolverParams.method = 'LRBFGS';
    end
    
   
    % compute the Karcher mean
    [Xopt, f, gf, gfgf0, iter, nf, ng, nR, nV, nVp, nH, ComTime, funs, grads, times, ...
        diss, initstepsizes, stepsizes, truestepsizes, relfs] = TestSPDMean(Ls, Xtrue, Xinitial, HasHHR, SolverParams);
    SPDMean = reshape(Xopt.main, n, n);
end